//#include <windows.h>
#include <stdio.h>
#include "shader.h"


void vector_rotate(miVector * result, miVector * vect, miVector * axis, miScalar angle)
{
	miVector v1,v2,v3;
	miScalar cosangle;

	//  .     3 :
	//Va -   .
	//Vapx -         ( )
	//Vapy -    ,  Vapx.
	// Va=dot(Ax,Vect)
	// Vapx=Vect-Va
	// Vapy=cross(Ax,Vect)
	//        :
	// Vplane=sin(a)*Vapy + cos(a)*Vapx
	//        Va, . . Res=Va+sin(a)*Vapy+cos(a)*Vapx.
	//    : Res= sin(a)*cross(Ax,Vect)+Ax*dot(Ax,Vect)*(1-cos(a))+cos(a)*Vect;
	cosangle=cos(angle);
	v1=*axis;
	mi_vector_normalize(&v1);
	v2=v1;
	mi_vector_mul(&v2,mi_vector_dot(&v1,vect)*(((miScalar)1.0)-cosangle));	//v2=Ax*dot(Ax,Vect)*(1-cos(a))
	mi_vector_prod(&v3,&v1,vect);
	mi_vector_mul(&v3,sin(angle));	//v3=cross(Ax,Vect)*sin(a)
	*result=*vect;
	mi_vector_mul(result,cosangle);	//result=Vect*cos(a)
	mi_vector_add(&v1,result,&v2);
	mi_vector_add(result,&v1,&v3);
}

extern "C" DLLEXPORT int mfisheye_a_version() {return 1;}

struct lensparam {
	miScalar HorizontalAngle;
	miScalar HorizontalScale;
	miBoolean MaskLense;
	miBoolean SoftLenseEdge;
	miBoolean SoftFrameEdge;
	miScalar SoftEdgeSize;
};


extern "C" DLLEXPORT miBoolean mfisheye_a(
        miColor         *result,
        miState         *state,
        lensparam       *paras)
{
    miVector        camdir,v1,v2;
	miScalar x,y,t,Power,sfy,sfx,clolrScale,ses,scale;
	miBoolean MaskLense,SoftLenseEdge,SoftFrameEdge;

	MaskLense=*mi_eval_boolean(&(paras->MaskLense));
	SoftLenseEdge=*mi_eval_boolean(&(paras->SoftLenseEdge));
	if(SoftLenseEdge)MaskLense=miTRUE;

	mi_vector_to_camera(state, &camdir, &state->dir);
	scale=*mi_eval_scalar(&(paras->HorizontalScale));
	t = state->camera->focal / -camdir.z / (state->camera->aperture/2)/scale;
    x = t * camdir.x; //  X  -1  1 (  )
    y = t * camdir.y;// * state->camera->aspect;
	t=sqrt(x*x+y*y);
	if((t>1.0)&&(MaskLense))
	{
		result->r=0;
		result->g=0;
		result->b=0;
		result->a=0;
		return miTRUE;
	}
	SoftFrameEdge=*mi_eval_boolean(&(paras->SoftFrameEdge));
	clolrScale=(miScalar)1.0;
	if((SoftFrameEdge)||(SoftLenseEdge))ses=*mi_eval_scalar(&(paras->SoftEdgeSize));

	sfx=((miScalar)1.0)-t;
	if((SoftLenseEdge)&&(sfx<ses))clolrScale*=sfx/ses;

	if(SoftFrameEdge)
	{
		sfy=y*state->camera->aspect*scale;
		if(sfy<0)sfy=((miScalar)1.0)+sfy;else sfy=((miScalar)1.0)-sfy;
		if(x<0)sfx=((miScalar)1.0)+x*scale; else sfx=((miScalar)1.0)-x*scale;
		if(sfy<ses)clolrScale*=sfy/ses;
		if(sfx<ses)clolrScale*=sfx/ses;
	}

	Power=*mi_eval_scalar(&(paras->HorizontalAngle))/((miScalar)(360.0/3.1415926535897932384626433832795));

	v1.x=-y;
	v1.y=x;
	v1.z=0;
	v2.x=0;
	v2.y=0;
	v2.z=-1;
	vector_rotate(&camdir,&v2,&v1,-t*Power);

	mi_vector_from_camera(state, &camdir, &camdir);
	MaskLense=mi_trace_eye(result, state, &state->org, &camdir);
	result->r*=clolrScale;
	result->g*=clolrScale;
	result->b*=clolrScale;
	result->a*=clolrScale;
	return(MaskLense);
}



extern "C" DLLEXPORT int environmentLense_version() {return 1;}


extern "C" DLLEXPORT miBoolean environmentLense(
        miColor         *result,
        miState         *state,
        void       *paras)
{
    miVector        camdir,v1,v2;
	miScalar x,y,t;

	mi_vector_to_camera(state, &camdir, &state->dir);
	t = state->camera->focal / -camdir.z / (state->camera->aperture/2)*((miScalar)(3.1415926535897932384626433832795));
    x = t * camdir.x;
    y = t * camdir.y * state->camera->aspect;
	camdir.x=0;
	camdir.y=0;
	camdir.z=-1;
	v2.x=1;
	v2.y=0;
	v2.z=0;
	vector_rotate(&v1,&camdir,&v2,y/2.0);
	v2.x=0;
	v2.y=1;
	vector_rotate(&camdir,&v1,&v2,x);
	mi_vector_from_camera(state, &camdir, &camdir);
	return(mi_trace_eye(result, state, &state->org, &camdir));
}



